@@ -5,21 +5,7 @@
 """
 IMPORTANT NOTE!
 
-To use this module on Mac OS X, you need the PyObjC module installed.
-For Python 3, run:
-    sudo pip3 install pyobjc-core
-    sudo pip3 install pyobjc
-For Python 2, run:
-    sudo pip install pyobjc-core
-    sudo pip install pyobjc
-(There's some bug with their installer, so install pyobjc-core first or else
-the install takes forever.)
 
-To use this module on Linux, you need Xlib module installed.
-For Python 3, run:
-    sudo pip3 install python3-Xlib
-For Python 2, run:
-    sudo pip install Xlib
 
 To use this module on Windows, you do not need anything else.
 
@@ -30,10 +16,9 @@ You will need PIL/Pillow to use the screenshot features.
 from __future__ import absolute_import, division, print_function
 
 
-__version__ = '0.9.44'
+__version__ = '0.9.45'
 
-import sys
-import time
+import sys, time, datetime, os, platform
 
 
 if sys.version_info[0] == 2 or sys.version_info[0:2] in ((3, 1), (3,2)):
@@ -134,6 +119,12 @@ KEY_NAMES = ['\t', '\n', '\r', ' ', '!', '"', '#', '$', '%', '&', "'", '(',
      'command', 'option', 'optionleft', 'optionright']
 KEYBOARD_KEYS = KEY_NAMES   # keeping old KEYBOARD_KEYS for backwards compatibility
 
+# Constants for the mouse button names:
+LEFT      = 'left'
+MIDDLE    = 'middle'
+RIGHT     = 'right'
+PRIMARY   = 'primary'
+SECONDARY = 'secondary'
 
 def isShiftCharacter(character):
     """Returns True if the key character is uppercase or shifted."""
@@ -148,9 +139,10 @@ elif sys.platform == 'darwin':
     from . import _pyautogui_osx as platformModule
 elif sys.platform == 'win32':
     from . import _pyautogui_win as platformModule
-else:
+elif platform.system() == 'Linux':
     from . import _pyautogui_x11 as platformModule
-
+else:
+    raise NotImplementedError('Your platform (%s) is not supported by PyAutoGUI.' % (platform.system()))
 
 # TODO: Having module-wide user-writable global variables is bad. It makes
 # restructuring the code very difficult. For instance, what if we decide to
@@ -165,10 +157,18 @@ MINIMUM_DURATION = 0.1
 # TODO: This value should vary with the platform. http://stackoverflow.com/q/1133857
 MINIMUM_SLEEP = 0.05
 PAUSE = 0.1 # The number of seconds to pause after EVERY public function call. Useful for debugging.
-FAILSAFE = True
-FAILSAFE_POINT = (0, 0) # If the mouse is here and FAILSAFE is True, the FailSafeException is raised.
 
 
+# If the mouse is over a coordinate in FAILSAFE_POINTS and FAILSAFE is True, the FailSafeException is raised.
+# The rest of the points are added to the FAILSAFE_POINTS list at the bottom of this file, after size() has been defined.
+# The points are for the corners of the screen, but note that these points don't automatically change if the screen resolution changes.
+FAILSAFE = True
+FAILSAFE_POINTS = [(0, 0)]
+
+LOG_SCREENSHOTS = False # If True, save screenshots for clicks and key presses.
+LOG_SCREENSHOTS_LIMIT = 10 # If not None, PyAutoGUI deletes old screenshots when this limit has been reached.
+G_LOG_SCREENSHOTS_FILENAMES = [] # TODO - make this a deque
+
 Point = collections.namedtuple('Point', 'x y')
 Size = collections.namedtuple('Size', 'width height')
 
@@ -200,11 +200,19 @@ def linear(n):
 
 
 def _autoPause(pause, _pause):
-    if _pause:
-        if pause is not None:
-            time.sleep(pause)
-        elif PAUSE != 0:
-            time.sleep(PAUSE)
+    """If `pause` is not `None`, then sleep for `pause` seconds.
+    If `_pause` is `True`, then sleep for `PAUSE` seconds (the global pause setting).
+
+    This function is called at the end of all of PyAutoGUI's mouse and keyboard functions. Normally, `_pause`
+    is set to `True` to add a short sleep so that the user can engage the failsafe. By default, this sleep
+    is as long as `PAUSE` settings. However, this can be override by setting `pause`, in which case the sleep
+    is as long as `pause` seconds.
+    """
+    if pause is not None:
+        time.sleep(pause)
+    elif _pause:
+        assert isinstance(PAUSE, int) or isinstance(PAUSE, float)
+        time.sleep(PAUSE)
 
 
 def _unpackXY(x, y):
@@ -238,6 +246,31 @@ def _unpackXY(x, y):
     return x, y
 
 
+def _logScreenshot(logScreenshot, action, specifics, folder='.'):
+    if logScreenshot == False:
+        return # Don't take a screenshot.
+    if logScreenshot is None and LOG_SCREENSHOTS == False:
+        return # Don't take a screenshot.
+
+    # Ensure that the "specifics" string isn't too long for the filename:
+    if len(specifics) > 12:
+        specifics = specifics[:12] + '...'
+
+    now = datetime.datetime.now()
+    filename = '%s-%s-%s_%s-%s-%s-%s_%s_%s.png' % (now.year, str(now.month).rjust(2, '0'), str(now.day).rjust(2, '0'),
+                                                      now.hour, now.minute, now.second, str(now.microsecond)[:3],
+                                                      action, specifics)
+    filepath = os.path.join(folder, filename)
+
+    # Delete the oldest screenshot if we've reached the maximum:
+    if (LOG_SCREENSHOTS_LIMIT is not None) and (len(G_LOG_SCREENSHOTS_FILENAMES) >= LOG_SCREENSHOTS_LIMIT):
+        os.unlink(os.path.join(folder, G_LOG_SCREENSHOTS_FILENAMES[0]))
+        del G_LOG_SCREENSHOTS_FILENAMES[0]
+
+    screenshot(filepath)
+    G_LOG_SCREENSHOTS_FILENAMES.append(filename)
+
+
 def position(x=None, y=None):
     """Returns the current xy coordinates of the mouse cursor as a two-integer
     tuple.
@@ -250,6 +283,8 @@ def position(x=None, y=None):
 
     Returns:
       (x, y) tuple of the current xy coordinates of the mouse cursor.
+
+    NOTE: The position() functon doesn't check for failsafe.
     """
     posx, posy = platformModule._position()
     posx = int(posx)
@@ -294,7 +329,65 @@ def onScreen(x, y=None):
 # Mouse Functions
 # ===============
 
-def mouseDown(x=None, y=None, button='left', duration=0.0, tween=linear, pause=None, _pause=True):
+"""
+NOTE: Although "mouse1" and "mouse2" buttons usually refer to the left and
+right mouse buttons respectively, in PyAutoGUI 1, 2, and 3 refer to the left,
+middle, and right buttons, respectively. This is because Xlib interprets
+button 2 as the middle button and button 3 as the right button, so we hold
+that for Windows and macOS as well (since those operating systems don't use
+button numbers but rather just "left" or "right").
+"""
+
+def _translateButton(button):
+    """
+    The left, middle, and right mouse buttons are button numbers 1, 2, and 3
+    respectively. This is the numbering that Xlib on Linux uses (while Windows
+    and macOS don't care about numbers; they just use "left" and "right").
+
+    The 'left' and 'right' mouse buttons will always refer to the physical
+    buttons on the mouse. The same applies for button 1 and 3.
+
+    However, if `button` is 'primary' or 'secondary', then we must check if
+    the mouse buttons have been "swapped" by the operating system's mouse
+    settings. If not swapped, the primary and secondary buttons are the left
+    and right mouse buttons respectively, and if swapped, the primary and
+    secondary buttons are the right and left mouse buttons, respectively.
+
+    TODO - The swap detection hasn't been done yet.
+    """
+    # TODO - We should check the OS settings to see if it's a left-hand setup, where button 1 would be "right".
+
+    # Check that `button` has a valid value:
+    button = button.lower()
+    if platform.system() == 'Linux':
+        # Check for valid button arg on Linux:
+        if button not in (LEFT, MIDDLE, RIGHT, PRIMARY, SECONDARY, 1, 2, 3, 4, 5, 6, 7):
+            raise ValueError("button argument must be one of ('left', 'middle', 'right', 'primary', 'secondary', 1, 2, 3, 4, 5, 6, 7)")
+    else:
+        # Check for valid button arg on Windows and macOS:
+        if button not in (LEFT, MIDDLE, RIGHT, PRIMARY, SECONDARY, 1, 2, 3):
+            raise ValueError("button argument must be one of ('left', 'middle', 'right', 'primary', 'secondary', 1, 2, 3)")
+
+    # TODO - Check if the primary/secondary mouse buttons have been swapped:
+    if button in (PRIMARY, SECONDARY):
+        swapped = False # TODO - Add the operating system-specific code to detect mouse swap later.
+        if swapped:
+            if button == PRIMARY:
+                return RIGHT
+            elif button == SECONDARY:
+                return LEFT
+        else:
+            if button == PRIMARY:
+                return LEFT
+            elif button == SECONDARY:
+                return RIGHT
+
+    # Return a mouse button integer value, not a string like 'left':
+    return {LEFT: LEFT, MIDDLE: MIDDLE, RIGHT: RIGHT,
+            1: LEFT, 2: MIDDLE, 3:RIGHT, 4:4, 5:5, 6:6, 7:7}[button]
+
+
+def mouseDown(x=None, y=None, button=PRIMARY, duration=0.0, tween=linear, pause=None, logScreenshot=None, _pause=True):
     """Performs pressing a mouse button down (but not up).
 
     The x and y parameters detail where the mouse event happens. If None, the
@@ -309,9 +402,7 @@ def mouseDown(x=None, y=None, button='left', duration=0.0, tween=linear, pause=N
         the screen with locateOnScreen() and click the center of.
       y (int, float, None, optional): The y position on the screen where the
         mouse down happens. None by default.
-      button (str, int, optional): The mouse button pressed down. Must be one of
-        'left', 'middle', 'right' (or 1, 2, or 3) respectively. 'left' by
-        default.
+      button (str, int, optional): The mouse button pressed down. TODO
 
     Returns:
       None
@@ -319,26 +410,20 @@ def mouseDown(x=None, y=None, button='left', duration=0.0, tween=linear, pause=N
     Raises:
       ValueError: If button is not one of 'left', 'middle', 'right', 1, 2, or 3
     """
-    if button not in ('left', 'middle', 'right', 1, 2, 3):
-        raise ValueError("button argument must be one of ('left', 'middle', 'right', 1, 2, 3), not %s" % button)
-
+    button = _translateButton(button)
     _failSafeCheck()
     x, y = _unpackXY(x, y)
 
     _mouseMoveDrag('move', x, y, 0, 0, duration=0, tween=None)
 
-    x, y = platformModule._position() # TODO - this isn't right. We need to check the params.
-    if button == 1 or str(button).lower() == 'left':
-        platformModule._mouseDown(x, y, 'left')
-    elif button == 2 or str(button).lower() == 'middle':
-        platformModule._mouseDown(x, y, 'middle')
-    elif button == 3 or str(button).lower() == 'right':
-        platformModule._mouseDown(x, y, 'right')
+    x, y = platformModule._position() # TODO Why do we call _position() here and overwrite x, y?
+    _logScreenshot(logScreenshot, 'mouseDown', '%s,%s' % (x, y), folder='.')
+    platformModule._mouseDown(x, y, button)
 
     _autoPause(pause, _pause)
 
 
-def mouseUp(x=None, y=None, button='left', duration=0.0, tween=linear, pause=None, _pause=True):
+def mouseUp(x=None, y=None, button=PRIMARY, duration=0.0, tween=linear, pause=None, logScreenshot=None, _pause=True):
     """Performs releasing a mouse button up (but not down beforehand).
 
     The x and y parameters detail where the mouse event happens. If None, the
@@ -353,9 +438,7 @@ def mouseUp(x=None, y=None, button='left', duration=0.0, tween=linear, pause=Non
         the screen with locateOnScreen() and click the center of.
       y (int, float, None, optional): The y position on the screen where the
         mouse up happens. None by default.
-      button (str, int, optional): The mouse button released. Must be one of
-        'left', 'middle', 'right' (or 1, 2, or 3) respectively. 'left' by
-        default.
+      button (str, int, optional): The mouse button released. TODO
 
     Returns:
       None
@@ -363,25 +446,20 @@ def mouseUp(x=None, y=None, button='left', duration=0.0, tween=linear, pause=Non
     Raises:
       ValueError: If button is not one of 'left', 'middle', 'right', 1, 2, or 3
     """
-    if button not in ('left', 'middle', 'right', 1, 2, 3):
-        raise ValueError("button argument must be one of ('left', 'middle', 'right', 1, 2, 3), not %s" % button)
-
+    button = _translateButton(button)
     _failSafeCheck()
     x, y = _unpackXY(x, y)
 
     _mouseMoveDrag('move', x, y, 0, 0, duration=0, tween=None)
 
-    x, y = platformModule._position()
-    if button == 1 or str(button).lower() == 'left':
-        platformModule._mouseUp(x, y, 'left')
-    elif button == 2 or str(button).lower() == 'middle':
-        platformModule._mouseUp(x, y, 'middle')
-    elif button == 3 or str(button).lower() == 'right':
-        platformModule._mouseUp(x, y, 'right')
+    x, y = platformModule._position() # TODO Why do we call _position() here and overwrite x, y?
+    _logScreenshot(logScreenshot, 'mouseUp', '%s,%s' % (x, y), folder='.')
+    platformModule._mouseUp(x, y, button)
 
     _autoPause(pause, _pause)
 
-def click(x=None, y=None, clicks=1, interval=0.0, button='left', duration=0.0, tween=linear, pause=None, _pause=True):
+
+def click(x=None, y=None, clicks=1, interval=0.0, button=PRIMARY, duration=0.0, tween=linear, pause=None, logScreenshot=None, _pause=True):
     """Performs pressing a mouse button down and then immediately releasing it.
 
     The x and y parameters detail where the mouse event happens. If None, the
@@ -401,9 +479,7 @@ def click(x=None, y=None, clicks=1, interval=0.0, button='left', duration=0.0, t
       interval (float, optional): The number of seconds in between each click,
         if the number of clicks is greater than 1. 0.0 by default, for no
         pause in between clicks.
-      button (str, int, optional): The mouse button clicked. Must be one of
-        'left', 'middle', 'right' (or 1, 2, or 3) respectively. 'left' by
-        default.
+      button (str, int, optional): The mouse button clicked. TODO
 
     Returns:
       None
@@ -411,32 +487,29 @@ def click(x=None, y=None, clicks=1, interval=0.0, button='left', duration=0.0, t
     Raises:
       ValueError: If button is not one of 'left', 'middle', 'right', 1, 2, 3
     """
-    if button not in ('left', 'middle', 'right', 1, 2, 3):
-        raise ValueError("button argument must be one of ('left', 'middle', 'right', 1, 2, 3)")
-
+    button = _translateButton(button)
     _failSafeCheck()
     x, y = _unpackXY(x, y)
 
     _mouseMoveDrag('move', x, y, 0, 0, duration, tween)
 
-    x, y = platformModule._position()
-    for i in range(clicks):
-        _failSafeCheck()
-        if button == 1 or str(button).lower() == 'left':
-            platformModule._click(x, y, 'left')
-        elif button == 2 or str(button).lower() == 'middle':
-            platformModule._click(x, y, 'middle')
-        elif button == 3 or str(button).lower() == 'right':
-            platformModule._click(x, y, 'right')
-        else:
-            # These mouse buttons for hor. and vert. scrolling only apply to x11:
-            platformModule._click(x, y, button)
+    x, y = platformModule._position() # TODO Why do we call _position() here and overwrite x, y?
 
-        time.sleep(interval)
+    _logScreenshot(logScreenshot, 'click', '%s,%s,%s,%s' % (button, clicks, x, y), folder='.')
+
+    if sys.platform == 'darwin':
+        platformModule._multiClick(x, y, button, 3)
+    else:
+        for i in range(clicks):
+            _failSafeCheck()
+            if button in (LEFT, MIDDLE, RIGHT):
+                platformModule._click(x, y, button)
+
+            time.sleep(interval)
 
     _autoPause(pause, _pause)
 
-def rightClick(x=None, y=None, duration=0.0, tween=linear, pause=None, _pause=True):
+def leftClick(x=None, y=None, interval=0.0, duration=0.0, tween=linear, pause=None, logScreenshot=None, _pause=True):
     """Performs a right mouse button click.
 
     This is a wrapper function for click('right', x, y).
@@ -453,18 +526,48 @@ def rightClick(x=None, y=None, duration=0.0, tween=linear, pause=None, _pause=Tr
         the screen with locateOnScreen() and click the center of.
       y (int, float, None, optional): The y position on the screen where the
         click happens. None by default.
+      interval (float, optional): The number of seconds in between each click,
+        if the number of clicks is greater than 1. 0.0 by default, for no
+        pause in between clicks.
 
     Returns:
       None
     """
     _failSafeCheck()
-
-    click(x, y, 1, 0.0, 'right', _pause=False)
-
+    click(x, y, 1, interval, LEFT, duration, tween, pause, logScreenshot, _pause=_pause)
     _autoPause(pause, _pause)
 
 
-def middleClick(x=None, y=None, duration=0.0, tween=linear, pause=None, _pause=True):
+def rightClick(x=None, y=None, interval=0.0, duration=0.0, tween=linear, pause=None, logScreenshot=None, _pause=True):
+    """Performs a right mouse button click.
+
+    This is a wrapper function for click('right', x, y).
+
+    The x and y parameters detail where the mouse event happens. If None, the
+    current mouse position is used. If a float value, it is rounded down. If
+    outside the boundaries of the screen, the event happens at edge of the
+    screen.
+
+    Args:
+      x (int, float, None, tuple, optional): The x position on the screen where the
+        click happens. None by default. If tuple, this is used for x and y.
+        If x is a str, it's considered a filename of an image to find on
+        the screen with locateOnScreen() and click the center of.
+      y (int, float, None, optional): The y position on the screen where the
+        click happens. None by default.
+      interval (float, optional): The number of seconds in between each click,
+        if the number of clicks is greater than 1. 0.0 by default, for no
+        pause in between clicks.
+
+    Returns:
+      None
+    """
+    _failSafeCheck()
+    click(x, y, 1, interval, RIGHT, duration, tween, pause, logScreenshot, _pause=_pause)
+    _autoPause(pause, _pause)
+
+
+def middleClick(x=None, y=None, interval=0.0, duration=0.0, tween=linear, pause=None, logScreenshot=None, _pause=True):
     """Performs a middle mouse button click.
 
     This is a wrapper function for click('right', x, y).
@@ -486,13 +589,11 @@ def middleClick(x=None, y=None, duration=0.0, tween=linear, pause=None, _pause=T
       None
     """
     _failSafeCheck()
-
-    click(x, y, 1, 0.0, 'middle', _pause=False)
-
+    click(x, y, 1, interval, MIDDLE, duration, tween, pause, logScreenshot, _pause=_pause)
     _autoPause(pause, _pause)
 
 
-def doubleClick(x=None, y=None, interval=0.0, button='left', duration=0.0, tween=linear, pause=None, _pause=True):
+def doubleClick(x=None, y=None, interval=0.0, button=LEFT, duration=0.0, tween=linear, pause=None, logScreenshot=None, _pause=True):
     """Performs a double click.
 
     This is a wrapper function for click('left', x, y, 2, interval).
@@ -512,9 +613,7 @@ def doubleClick(x=None, y=None, interval=0.0, button='left', duration=0.0, tween
       interval (float, optional): The number of seconds in between each click,
         if the number of clicks is greater than 1. 0.0 by default, for no
         pause in between clicks.
-      button (str, int, optional): The mouse button clicked. Must be one of
-        'left', 'middle', 'right' (or 1, 2, or 3) respectively. 'left' by
-        default.
+      button (str, int, optional): The mouse button released. TODO
 
     Returns:
       None
@@ -530,14 +629,16 @@ def doubleClick(x=None, y=None, interval=0.0, button='left', duration=0.0, tween
         x, y = _unpackXY(x, y)
         _mouseMoveDrag('move', x, y, 0, 0, duration=0, tween=None)
         x, y = platformModule._position()
+        _logScreenshot(logScreenshot, 'click', '%s,2,%s,%s' % (button, x, y), folder='.')
         platformModule._multiClick(x, y, button, 2)
     else:
-        click(x, y, 2, interval, button, _pause=False)
+		# Click for Windows or Linux:
+        click(x, y, 2, interval, button, duration, tween, pause, logScreenshot, _pause=False)
 
     _autoPause(pause, _pause)
 
 
-def tripleClick(x=None, y=None, interval=0.0, button='left', duration=0.0, tween=linear, pause=None, _pause=True):
+def tripleClick(x=None, y=None, interval=0.0, button=LEFT, duration=0.0, tween=linear, pause=None, logScreenshot=None, _pause=True):
     """Performs a triple click..
 
     This is a wrapper function for click('left', x, y, 3, interval).
@@ -557,9 +658,7 @@ def tripleClick(x=None, y=None, interval=0.0, button='left', duration=0.0, tween
       interval (float, optional): The number of seconds in between each click,
         if the number of clicks is greater than 1. 0.0 by default, for no
         pause in between clicks.
-      button (str, int, optional): The mouse button clicked. Must be one of
-        'left', 'middle', 'right' (or 1, 2, or 3) respectively. 'left' by
-        default.
+      button (str, int, optional): The mouse button released. TODO
 
     Returns:
       None
@@ -575,13 +674,15 @@ def tripleClick(x=None, y=None, interval=0.0, button='left', duration=0.0, tween
         x, y = _unpackXY(x, y)
         _mouseMoveDrag('move', x, y, 0, 0, duration=0, tween=None)
         x, y = platformModule._position()
+        _logScreenshot(logScreenshot, 'click', '%s,3,%s,%s' % (x, y), folder='.')
         platformModule._multiClick(x, y, button, 3)
     else:
-        click(x, y, 2, interval, button, _pause=False)
+		# Click for Windows or Linux:
+        click(x, y, 3, interval, button, duration, tween, pause, logScreenshot, _pause=False)
     _autoPause(pause, _pause)
 
 
-def scroll(clicks, x=None, y=None, pause=None, _pause=True):
+def scroll(clicks, x=None, y=None, pause=None, logScreenshot=None, _pause=True):
     """Performs a scroll of the mouse scroll wheel.
 
     Whether this is a vertical or horizontal scroll depends on the underlying
@@ -607,12 +708,13 @@ def scroll(clicks, x=None, y=None, pause=None, _pause=True):
         x, y = x[0], x[1]
     x, y = position(x, y)
 
+    _logScreenshot(logScreenshot, 'scroll', '%s,%s,%s' % (clicks,x, y), folder='.')
     platformModule._scroll(clicks, x, y)
 
     _autoPause(pause, _pause)
 
 
-def hscroll(clicks, x=None, y=None, pause=None, _pause=True):
+def hscroll(clicks, x=None, y=None, pause=None, logScreenshot=None, _pause=True):
     """Performs an explicitly horizontal scroll of the mouse scroll wheel,
     if this is supported by the operating system. (Currently just Linux.)
 
@@ -636,12 +738,13 @@ def hscroll(clicks, x=None, y=None, pause=None, _pause=True):
         x, y = x[0], x[1]
     x, y = position(x, y)
 
+    _logScreenshot(logScreenshot, 'hscroll', '%s,%s,%s' % (clicks, x, y), folder='.')
     platformModule._hscroll(clicks, x, y)
 
     _autoPause(pause, _pause)
 
 
-def vscroll(clicks, x=None, y=None, pause=None, _pause=True):
+def vscroll(clicks, x=None, y=None, pause=None, logScreenshot=None, _pause=True):
     """Performs an explicitly vertical scroll of the mouse scroll wheel,
     if this is supported by the operating system. (Currently just Linux.)
 
@@ -664,12 +767,14 @@ def vscroll(clicks, x=None, y=None, pause=None, _pause=True):
     if type(x) in (tuple, list):
         x, y = x[0], x[1]
     x, y = position(x, y)
+
+    _logScreenshot(logScreenshot, 'vscroll', '%s,%s,%s' % (clicks, x, y), folder='.')
     platformModule._vscroll(clicks, x, y)
 
     _autoPause(pause, _pause)
 
 
-def moveTo(x=None, y=None, duration=0.0, tween=linear, pause=None, _pause=True):
+def moveTo(x=None, y=None, duration=0.0, tween=linear, pause=None, logScreenshot=False, _pause=True):
     """Moves the mouse cursor to a point on the screen.
 
     The x and y parameters detail where the mouse event happens. If None, the
@@ -697,12 +802,13 @@ def moveTo(x=None, y=None, duration=0.0, tween=linear, pause=None, _pause=True):
     _failSafeCheck()
     x, y = _unpackXY(x, y)
 
+    _logScreenshot(logScreenshot, 'moveTo', '%s,%s' % (x, y), folder='.')
     _mouseMoveDrag('move', x, y, 0, 0, duration, tween)
 
     _autoPause(pause, _pause)
 
 
-def moveRel(xOffset=None, yOffset=None, duration=0.0, tween=linear, pause=None, _pause=True):
+def moveRel(xOffset=None, yOffset=None, duration=0.0, tween=linear, pause=None, logScreenshot=False, _pause=True):
     """Moves the mouse cursor to a point on the screen, relative to its current
     position.
 
@@ -731,6 +837,7 @@ def moveRel(xOffset=None, yOffset=None, duration=0.0, tween=linear, pause=None,
 
     xOffset, yOffset = _unpackXY(xOffset, yOffset)
 
+    _logScreenshot(logScreenshot, 'moveRel', '%s,%s' % (xOffset, yOffset), folder='.')
     _mouseMoveDrag('move', None, None, xOffset, yOffset, duration, tween)
 
     _autoPause(pause, _pause)
@@ -739,7 +846,7 @@ def moveRel(xOffset=None, yOffset=None, duration=0.0, tween=linear, pause=None,
 move = moveRel # For PyAutoGUI 1.0, move() replaces moveRel().
 
 
-def dragTo(x=None, y=None, duration=0.0, tween=linear, button='left', pause=None, _pause=True, mouseDownUp=True):
+def dragTo(x=None, y=None, duration=0.0, tween=linear, button=PRIMARY, pause=None, logScreenshot=None, _pause=True, mouseDownUp=True):
     """Performs a mouse drag (mouse movement while a button is held down) to a
     point on the screen.
 
@@ -761,9 +868,7 @@ def dragTo(x=None, y=None, duration=0.0, tween=linear, button='left', pause=None
       tween (func, optional): The tweening function used if the duration is not
         0. A linear tween is used by default. See the tweens.py file for
         details.
-      button (str, int, optional): The mouse button clicked. Must be one of
-        'left', 'middle', 'right' (or 1, 2, or 3) respectively. 'left' by
-        default.
+      button (str, int, optional): The mouse button released. TODO
       mouseDownUp (True, False): When true, the mouseUp/Down actions are not perfomed.
         Which allows dragging over multiple (small) actions. 'True' by default.
 
@@ -773,16 +878,17 @@ def dragTo(x=None, y=None, duration=0.0, tween=linear, button='left', pause=None
     _failSafeCheck()
     x, y = _unpackXY(x, y)
 
+    _logScreenshot(logScreenshot, 'dragTo', '%s,%s' % (x, y), folder='.')
     if mouseDownUp:
-        mouseDown(button=button, _pause=False)
+        mouseDown(button=button, logScreenshot=False, _pause=False)
     _mouseMoveDrag('drag', x, y, 0, 0, duration, tween, button)
     if mouseDownUp:
-        mouseUp(button=button, _pause=False)
+        mouseUp(button=button, logScreenshot=False, _pause=False)
 
     _autoPause(pause, _pause)
 
 
-def dragRel(xOffset=0, yOffset=0, duration=0.0, tween=linear, button='left', pause=None, _pause=True, mouseDownUp=True):
+def dragRel(xOffset=0, yOffset=0, duration=0.0, tween=linear, button=PRIMARY, pause=None, logScreenshot=None, _pause=True, mouseDownUp=True):
     """Performs a mouse drag (mouse movement while a button is held down) to a
     point on the screen, relative to its current position.
 
@@ -802,9 +908,7 @@ def dragRel(xOffset=0, yOffset=0, duration=0.0, tween=linear, button='left', pau
       tween (func, optional): The tweening function used if the duration is not
         0. A linear tween is used by default. See the tweens.py file for
         details.
-      button (str, int, optional): The mouse button clicked. Must be one of
-        'left', 'middle', 'right' (or 1, 2, or 3) respectively. 'left' by
-        default.
+      button (str, int, optional): The mouse button released. TODO
       mouseDownUp (True, False): When true, the mouseUp/Down actions are not perfomed.
         Which allows dragging over multiple (small) actions. 'True' by default.
 
@@ -825,11 +929,12 @@ def dragRel(xOffset=0, yOffset=0, duration=0.0, tween=linear, button='left', pau
     _failSafeCheck()
 
     mousex, mousey = platformModule._position()
+    _logScreenshot(logScreenshot, 'dragRel', '%s,%s' % (xOffset, yOffset), folder='.')
     if mouseDownUp:
-        mouseDown(button=button, _pause=False)
+        mouseDown(button=button, logScreenshot=False, _pause=False)
     _mouseMoveDrag('drag', mousex, mousey, xOffset, yOffset, duration, tween, button)
     if mouseDownUp:
-        mouseUp(button=button, _pause=False)
+        mouseUp(button=button, logScreenshot=False, _pause=False)
 
     _autoPause(pause, _pause)
 
@@ -864,9 +969,7 @@ def _mouseMoveDrag(moveOrDrag, x, y, xOffset, yOffset, duration, tween=linear, b
       tween (func, optional): The tweening function used if the duration is not
         0. A linear tween is used by default. See the tweens.py file for
         details.
-      button (str, int, optional): The mouse button clicked. Must be one of
-        'left', 'middle', 'right' (or 1, 2, or 3) respectively. 'left' by
-        default.
+      button (str, int, optional): The mouse button released. TODO
 
     Returns:
       None
@@ -897,8 +1000,8 @@ def _mouseMoveDrag(moveOrDrag, x, y, xOffset, yOffset, duration, tween=linear, b
     width, height = size()
 
     # Make sure x and y are within the screen bounds.
-    x = max(0, min(x, width - 1))
-    y = max(0, min(y, height - 1))
+    #x = max(0, min(x, width - 1))
+    #y = max(0, min(y, height - 1))
 
     # If the duration is small enough, just move the cursor there instantly.
     steps = [(x, y)]
@@ -959,7 +1062,7 @@ def isValidKey(key):
     return platformModule.keyboardMapping.get(key, None) != None
 
 
-def keyDown(key, pause=None, _pause=True):
+def keyDown(key, pause=None, logScreenshot=None, _pause=True):
     """Performs a keyboard key press without the release. This will put that
     key in a held down state.
 
@@ -977,11 +1080,12 @@ def keyDown(key, pause=None, _pause=True):
         key = key.lower()
 
     _failSafeCheck()
+    _logScreenshot(logScreenshot, 'keyDown', key, folder='.')
     platformModule._keyDown(key)
 
     _autoPause(pause, _pause)
 
-def keyUp(key, pause=None, _pause=True):
+def keyUp(key, pause=None, logScreenshot=None, _pause=True):
     """Performs a keyboard key release (without the press down beforehand).
 
     Args:
@@ -995,11 +1099,12 @@ def keyUp(key, pause=None, _pause=True):
         key = key.lower()
 
     _failSafeCheck()
+    _logScreenshot(logScreenshot, 'keyUp', key, folder='.')
     platformModule._keyUp(key)
 
     _autoPause(pause, _pause)
 
-def press(keys, presses=1, interval=0.0, pause=None, _pause=True):
+def press(keys, presses=1, interval=0.0, pause=None, logScreenshot=None, _pause=True):
     """Performs a keyboard key press down, followed by a release.
 
     Args:
@@ -1015,7 +1120,7 @@ def press(keys, presses=1, interval=0.0, pause=None, _pause=True):
       None
     """
     if type(keys) == str:
-        keys = [keys] # put string in a list
+        keys = [keys] # If keys is 'enter', convert it to ['enter'].
     else:
         lowerKeys = []
         for s in keys:
@@ -1024,6 +1129,7 @@ def press(keys, presses=1, interval=0.0, pause=None, _pause=True):
             else:
                 lowerKeys.append(s)
     interval = float(interval)
+    _logScreenshot(logScreenshot, 'press', ','.join(keys), folder='.')
     for i in range(presses):
         for k in keys:
             _failSafeCheck()
@@ -1033,7 +1139,7 @@ def press(keys, presses=1, interval=0.0, pause=None, _pause=True):
 
     _autoPause(pause, _pause)
 
-def typewrite(message, interval=0.0, pause=None, _pause=True):
+def typewrite(message, interval=0.0, pause=None, logScreenshot=None, _pause=True):
     """Performs a keyboard key press down, followed by a release, for each of
     the characters in message.
 
@@ -1054,10 +1160,11 @@ def typewrite(message, interval=0.0, pause=None, _pause=True):
     Returns:
       None
     """
-    interval = float(interval)
+    interval = float(interval) # TODO - this should be taken out.
 
     _failSafeCheck()
 
+    _logScreenshot(logScreenshot, 'write', message, folder='.')
     for c in message:
         if len(c) > 1:
             c = c.lower()
@@ -1087,10 +1194,11 @@ def hotkey(*args, **kwargs):
     Returns:
       None
     """
-    interval = float(kwargs.get('interval', 0.0))
+    interval = float(kwargs.get('interval', 0.0)) # TODO - this should be taken out.
 
     _failSafeCheck()
 
+    _logScreenshot(kwargs.get('logScreenshot'), 'hotkey', ','.join(args), folder='.')
     for c in args:
         if len(c) > 1:
             c = c.lower()
@@ -1110,23 +1218,8 @@ class FailSafeException(Exception):
 
 
 def _failSafeCheck():
-    global FAILSAFE_POINT
-
-    if isinstance(FAILSAFE_POINT, str):
-        if FAILSAFE_POINT.lower() == 'topleft':
-            FAILSAFE_POINT = (0, 0)
-        elif FAILSAFE_POINT.lower() == 'topright':
-            FAILSAFE_POINT = (size()[0] - 1, 0)
-        elif FAILSAFE_POINT.lower() == 'bottomleft':
-            FAILSAFE_POINT = (0, size()[1] - 1)
-        elif FAILSAFE_POINT.lower() == 'bottomright':
-            screenSize = size()
-            FAILSAFE_POINT = (screenSize[0] - 1, screenSize[1] - 1)
-        else:
-            raise
-
-    if FAILSAFE and position() == FAILSAFE_POINT:
-        raise FailSafeException('PyAutoGUI fail-safe triggered from mouse moving to upper-left corner. To disable this fail-safe, set pyautogui.FAILSAFE to False.')
+    if FAILSAFE and tuple(position()) in FAILSAFE_POINTS:
+        raise FailSafeException('PyAutoGUI fail-safe triggered from mouse moving to a corner of the screen. To disable this fail-safe, set pyautogui.FAILSAFE to False. DISABLING FAIL-SAFE IS NOT RECOMMENDED.')
 
 
 def displayMousePosition(xOffset=0, yOffset=0):
@@ -1144,7 +1237,7 @@ def displayMousePosition(xOffset=0, yOffset=0):
             if (x - xOffset) < 0 or (y - yOffset) < 0 or (x - xOffset) >= resolution[0] or (y - yOffset) >= resolution[1]:
                 pixelColor = ('NaN', 'NaN', 'NaN')
             else:
-                pixelColor = pyscreeze.screenshot().getpixel((x, y))
+                pixelColor = pyscreeze.screenshot().getpixel((x, y)) # NOTE: On Windows & Linux, getpixel() returns a 3-integer tuple, but on macOS it returns a 4-integer tuple.
             positionStr += ' RGB: (' + str(pixelColor[0]).rjust(3)
             positionStr += ', ' + str(pixelColor[1]).rjust(3)
             positionStr += ', ' + str(pixelColor[2]).rjust(3) + ')'
@@ -1154,3 +1247,8 @@ def displayMousePosition(xOffset=0, yOffset=0):
     except KeyboardInterrupt:
         sys.stdout.write('\n')
         sys.stdout.flush()
+
+
+# Add the bottom left, top right, and bottom right corners to FAILSAFE_POINTS.
+_right, _bottom = size()
+FAILSAFE_POINTS.extend([(0, _bottom - 1), (_right - 1, 0), (_right - 1, _bottom - 1)])
